{% extends "base.html" %}
{% set active_page = "index" %}

{% block contents %}
{% for msg in messages %}
<div class="alert alert-secondary">{{ msg }}</div>
{% endfor %}
{% if multiuser %}
<h1>Welcome {{ user.login }}</h1>
{% else %}
<h1>Welcome!</h1>
{% endif %}
<p>Here are your projects:</p>
<ul class="list-group">
{% for project in projects %}
  <li class="list-group-item"><a href="{{ reverse_url('project', project.id) }}">{{ project.name }}</a></li>
{% else %}
  <li class="list-group-item disabled" style="font-style: oblique;">You have no projects yet.</li>
{% endfor %}
  <li class="list-group-item list-group-item-primary" style="font-weight: bold;"><a href="{{ reverse_url('new_project') }}">Create a project</a></li>
</ul>
{% endblock %}
