{% extends "base.html" %}
{% set active_page = "index" %}

{% block contents %}
<div class="jumbotron">
  <h1 class="display-3">Welcome</h1>
  <p class="lead">The <span style="font-family:BadScript;"><a href="https://taguette.fr">TAGUETTE</a></span> project is an attempt to create an open-source web-based document tagging tool for qualitative data analysis.</p>
  <p>Using this tool, you can upload a collection of documents, create a hierarchy of tags, and annotate portions of documents with tags and notes that you can recall and organize later.</p>
  {% if register_enabled %}
  <p><a href="{{ reverse_url('register') }}">Register now</a> for free and get started!</p>
  {% endif %}
</div>
{% endblock %}
